# -*- coding: utf-8 -*-

"""Top-level package for Standalone HTML Interactive Phylogenetic Tree Viz."""

__author__ = """Peter Kruczkiewicz"""
__email__ = 'peter.kruczkiewicz@gmail.com'
__version__ = '0.4.0'
